
.include      "m8def.inc"



;**** DEFINITIONS **********************************************

;**** CONSTANTS

.equ  CLOCK       = 1000000                     ;Internal RC oscillator frequency in Hz

.equ  BAUD        = 9600                        ;UART baudrate for debugging and verifying
.equ  UBRRVAL     = CLOCK/(BAUD*16)-1

.equ  CRYSTAL     = 32786                       ;RTC crystal frequency in Hz

.equ  dZero       = 100                         ;DCF 0 impulse in ms
.equ  dOne        = 200                         ;DCF 1 impulse in ms
.equ  dTolerance  = 25                          ;DCF tolerance in ms

.equ  tTicksPS    = 128                         ;RTC timer ticks per second


;**** REGISTERS

;watch out for r26 and up if you should use X/Y ptr

.def  mulLow      = r0                          ;Multiplication result registers
.def  mulHigh     = r1

.def  sreg_backup = r2                          ;For a quick SREG backup without pushing and popping

.def  dLast       = r8                          ;DCF last interrupt time for calculating length of pulses

.def  dYear       = r9                          ;DCF decoded data
.def  dMonth      = r10
.def  dDay        = r11
.def  dHour       = r12
.def  dMinute     = r13
.def  dWeekday    = r15

.def  i           = r16                         ;General purpose working registers
.def  j           = r17

.def  tYear       = r19                         ;RTC
.def  tMonth      = r20
.def  tDay        = r21
.def  tHour       = r22
.def  tMinute     = r23
.def  tSecond     = r24
.def  tWeekday    = r25

.def  dPtr        = r26                         ;DCF data ptr
.def  dData       = r27                         ;DCF data
.def  dCrc        = r28                         ;DCF CRC


;***** FLAGS

.def  flags      = r29

.equ  tDst        = 0                           ;Daylight saving time; 0 = no; 1 = +1h 
.equ  tDstNext    = 1                           ;Change daylight saving time next hour

.equ  dDst        = 2                           ;DCF DST info
.equ  dDstNext    = 3

.equ	lOverride	= 4

.equ  sOne        = 6                           ;Simple timer for double click
.equ  sTwo        = 7


;***** SRAM

.dseg
.org 0x0101
;someBytes:  .byte 12


;**** INTERRUPT VECTORS

.cseg

.org  0x00
      rjmp  init
      
.org  INT0addr
      rjmp  iDcf

.org  INT1addr
      rjmp  iPower
      
.org  OC2addr
      rjmp  iTimer
      

;**** INITIALISATION ********************************************

init:

;**** STACK INIT

      ldi   i, LOW(RAMEND)                      ;Stackpointer start address
      out   SPL, i
      ldi   i, HIGH(RAMEND)
      out   SPH, i

;**** I/O & SLEEP INIT

	;PD0 DCF

      ldi   i, 0b00010001                       ;INT0/PD2 input, pull-up deactivated
      out   DDRD, i                             ;INT1 pull-up activated
      ldi   i, 0b00011000
      out   PortD, i
      
;      clr   i                                   ;PB1/2 LED
;      out   PortB, i
      ldi   i, 0b00000110
      out   DDRB, i

      ldi   i, (1<<ISC00)|(1<<SE)|(3<<SM0)      ;INT0 on change, INT1 on low level -> can wake up
      out   MCUCR, i
      ldi   i, (1<<INT0)|(1<<INT1)
      out   GICR, i

;**** Timer 2

      ldi   i, (1<<OCIE2)                       ;Use TCNT2 to measure up to 2 seconds
      out   TIMSK, i                            ;Needed for easy DCF pulse recognition
      ldi   i, (1<<AS2)                         ;Enable asynchronous mode
      out   ASSR, i
      ldi   i, (6<<CS20)                        ;Prescaler: 32768 Hz / 256 = 128 Hz
;      ldi   i, (1<<CS20)                       ;Faster clock for debugging
      out   TCCR2, i


;**** UART INITIALISIEREN
/*
      ldi   r16, LOW(UBRRVAL)                   ;USART for debugging
      out   UBRRL, r16
      ldi   r16, HIGH(UBRRVAL)
      out   UBRRH, r16
      ldi   r16, (1<<TXEN)
      out   UCSRB, r16
*/
;**** DEFAULT DATE
/*
      ldi   tYear,   15                         ;Init date and time for debugging
      ldi   tMonth,  12
      ldi   tDay,    31
      ldi   tHour,   21
      ldi   tMinute,  30
      ldi   tSecond,  55
      ldi   tWeekday, 7
      ldi   flags, 0;(1<<fDstNext)|(1<<fDst)
*/

;**** INTERRUPTS AKTIVIEREN

      sei                                       ;Global interrupt enable



;**** PROGRAM ***************************************************

;**** LOOP

loop:
      sbis  PinD, 3                             ;No SREG needed!
	rjmp  loop

      cbi	PortD, 0					;Turn off DCF

      ldi   j, (1<<INT0)|(1<<INT1)              ;Switch on wake up interrupt
      out   GICR, j

      sleep
      rjmp loop



;**** RTC *******************************************************

iTimer:
;      in    sreg_backup, SREG
      
      sbrc  flags, sOne					;Decrease double click timer
      cbr   flags, (1<<sOne)
      sbrc  flags, sTwo
      subi  flags, (1<<sOne)
	out 	UDR, flags
 
      inc   tSecond
      cpi   tSecond, 60                         ;Wrap seconds at 60
      brlo  iTimerEnd
      clr   tSecond

      inc   tMinute
      cpi   tMinute, 60                         ;Wrap minutes at 60
      brlo  iTimerEnd
      clr   tMinute

      sbrc  flags, tDstNext				;Change hour according to DST rules if tDstNext == 1
      rjmp  iTimerDst

      inc   tHour
      cpi   tWeekDay, 7                         ;Sunday?
      brne  iTimerHour
      cpi   tDay, 25                            ;Last sunday?
      brlo  iTimerHour
      cpi   tMonth, 3                           ;Last sunday in march?
      brne  iTimerDstOctober
      cpi   tHour, 1                            ;1 o'clock
      brne  iTimerHour
      sbr   flags, (1<<tDstNext)
;      rjmp  iTimerHour
iTimerDstOctober:
      cpi   tMonth, 10                          ;Last sunday in october?
      brne  iTimerHour
      cpi   tHour, 2
      brne  iTimerHour
      sbr   flags, (1<<tDstNext)
iTimerHour:

      cpi   tHour, 24                           ;Wrap hour at 24
      brlo  iTimerEnd
      clr   tHour

      inc   tWeekday
      cpi   tWeekday, 8                         ;Wrap weekday at 8
      brlo  iTimerDay
      ldi   tWeekday, 1

iTimerDay:
      inc   tDay
      cpi   tMonth, 2                           ;Watch out for leap day in febuary
      breq  iTimerFeb
      ldi   ZH, HIGH(tableMonth*2)
      mov   ZL, tMonth
      dec   ZL
      lpm   i, Z
      cp    tDay, i
      brlo  iTimerEnd
iTimerWrapDay:
      ldi   tDay, 1

      inc   tMonth
      cpi   tMonth, 12+1                        ;Wrap month at 13
      brlo  iTimerEnd
	ldi	tMonth, 1

      inc   tYear

iTimerEnd:
      in    i, OCR2
      subi  i, -tTicksPS
      out   OCR2, i
      
;      out   UDR, tSecond

	rcall	fLed

;      out   SREG, sreg_backup

      reti


iTimerDST:
      cbr   flags, (1<<tDstNext)

      cpi   tMonth, 10                          ;Last sunday in march?
      breq  iTimerDstWinter
      sbr   flags, (1<<tDst)

      ldi   tHour, 3

      rjmp  iTimerEnd

iTimerDstWinter:
      cbr   flags, (1<<tDst)

      ldi   tHour, 2

      rjmp  iTimerEnd
      

iTimerFeb:
      mov   i, tYear                            ;Neglecting /100 /400 leap day rule
      andi  i, 0b00000011                       ;Give me a call if this caused you any trouble
      breq  iTimerFeb29

      cpi   tDay, 28+1                          ;Wrap day after 28
;      brlo  iTimerEnd
;      rjmp  iTimerWrapDay
      rjmp  iTimerFeb28                         ;Saves 1 line
iTimerFeb29:
      cpi   tDay, 29+1                          ;Wrap day after 29
iTimerFeb28:
      brlo  iTimerEnd
      rjmp  iTimerWrapDay



;**** DCF *******************************************************

iDcf:
;      in    sreg_backup, SREG

      in    i, TCNT2

	push	i
      sub   i, dLast
	pop	dLast

      cpi   i, tTicksPS*(dZero-dTolerance)/1000
      brlo  iDcfEnd
      cpi   i, tTicksPS*(dZero+dTolerance)/1000
      brlo  iDcfZero
      cpi   i, tTicksPS*(dOne-dTolerance)/1000
      brlo  iDcfEnd
      cpi   i, tTicksPS*(dOne+dTolerance)/1000
      brlo  iDcfOne
      cpi   i, tTicksPS*(2000-dOne-dTolerance)/1000
      brlo  iDcfEnd
      cpi   i, tTicksPS*(2000-dZero+dTolerance)/1000
      brlo  iDcfStart
      
iDcfEnd:
;      out   SREG, sreg_backup

      reti

iDcfZero:
      clt                                       ;Clear carry bit for rolling

      rjmp  iDcfData

iDcfOne:
      cpi   dPtr, 16                            ;Save DST information
      brne  iDcfNotDstNext
      sbr   flags, (1<<dDstNext)
iDcfNotDstNext:
      cpi   dPtr, 17
      brne  iDcfNotDst
      sbr   flags, (1<<dDst)
iDcfNotDst:

      inc   dCRC
      andi  dCRC, 1                             ;Not the fastest way to do it, but saves a few lines later

      set                                       ;Set carry bit for rolling

iDcfData:
      inc   dPtr

      cpi   dPtr, 22                            ;With parity check, ignore last bit 
      brlo  iDcfCrcReset
      cpi   dPtr, 29
      breq  iDcfMin
      cpi   dPtr, 36
      breq  iDcfHour
      cpi   dPtr, 59
      breq  iDcfYear

      brtc  iDcfDataZero

      sec
      
iDcfDataPush:
      ror   dData

      cpi   dPtr, 42                            ;Without parity check, include last bit
      breq  iDcfDay
      cpi   dPtr, 45
      breq  iDcfWeekday
      cpi   dPtr, 50
      breq  iDcfMonth

      rjmp  iDcfEnd

iDcfDataZero:
      clc
      rjmp  iDcfDataPush
      

iDcfStart:
      cpi   dPtr, 59
      brlo  iDcfReset

      clr   tSecond                             ;Let's be accurate and reset the RTC
      out   TCNT2, tSecond
      ldi   i, (1<<OCF2)                        ;Clear interrupt flag by writing one
      out   TIFR, i
      ldi   i, tTicksPS                         ;Set next wakeup
      out   OCR2, i
;      clr   dLast                              ;First bit not needed anyway

;      sbis  PinD, 3					;Neglecting possibility that 1 fading step is left out
;	 rcall	fLed

      out   UDR, tSecond
      
      mov   tYear, dYear
      mov   tMonth, dMonth
      mov   tDay, dDay
      mov   tHour, dHour
      mov   tMinute, dMinute

      ori  	flags, 0b11111100
      sbrc  flags, dDstNext
      sbr   flags, (1<<tDstNext)
      sbrc  flags, dDst
      sbr   flags, (1<<tDst)
      
;      lsr   flags                              ;Works as long as there are no other flags
;      lsr   flags

iDcfReset:
      clr   dData
      clr   dPtr

      cbr   flags, (1<<dDstNext)|(1<<dDst)

      rjmp  iDcfEnd


iDcfMin:
      tst   dCrc                                ;CRC check
      brne  iDcfReset

      lsr   dData
      rcall iDcfBcd                             ;Convert and save received data
      mov   dMinute, r0
      
iDcfCrcReset:
      clr   dCrc
      rjmp  iDcfEnd


iDcfHour:
      tst   dCrc                                ;CRC check
      brne  iDcfReset

      lsr   dData
      lsr   dData
      rcall iDcfBcd                             ;Convert and save received data
      mov   dHour, r0

      rjmp  iDcfCrcReset


iDcfDay:
      lsr   dData
      lsr   dData
      rcall iDcfBcd                             ;Convert and save received data
      mov   dDay, r0

      rjmp  iDcfEnd


iDcfWeekday:
      lsr   dData
      swap  dData
      mov   dWeekday, dData

      clr   dData                               ;Usually in BCD to Binary

      rjmp  iDcfEnd


iDcfYear:
      tst   dCrc                                ;CRC check
      brne  iDcfReset

      rcall iDcfBcd                             ;Convert received data and apply new time information
      mov   dYear, r0                           ;We will be 1.8 - 1.9 seconds ahead
            
      rjmp  iDcfEnd


iDcfMonth:
      lsr   dData
      lsr   dData
      lsr   dData
      rcall iDcfBcd                             ;Convert and save received data
      mov   dMonth, r0

      rjmp  iDcfEnd


iDcfBcd:
	push	j
      mov   i, dData					

      andi  dData, 0b00001111				;Extract low nibble
      
      swap  i						;Extract high nibble and convert
      andi  i, 0b00001111
      ldi   j, 10
      mul   i, j

	pop	j

      add   r0, dData					;Add both numbers

;      out   UDR, r0

      clr   dData

      ret



;**** POWER ON INTERRUPT ****************************************

iPower:
      ldi   i, (1<<INT0)				;Clear INT1
      out   GICR, i

	sbi   PortD, 0                            ;Activate DCF

	cbr	flags, (1<<lOverride)
      cpi   flags, (1<<sOne)                    ;Is double click timer still running?
      sbr	flags, (1<<lOverride)
      cpi   flags, (1<<sTwo)
      sbr	flags, (1<<lOverride)

      ori   flags, (1<<sTwo)                    ;Enable double click for 1-2s

	rcall	fLed						;Activate PWM
      
      reti



;**** FUNCTIONS *************************************************

;**** SET LED

fLed:
      sbic  PinD, 3					;Activate LED only if power is on
	rjmp	fLedOff

	sec							;Set carry so brlo branches	
	sbrs	flags, lOverride
	rjmp	fLedPwmOff

	cpi	tHour, 22
	breq	fLedFade

fLedPwmOff:
	clr	i
	out	TCCR1A, i
	out	TCCR1B, i

	brlo	fLedDay					;earlier than 10?

fLedNight:	
	sbi	PortB, 1
	cbi	PortB, 2

	ret

fLedDay:
	cpi	tHour, 7					;earlier than 7?
	brlo	fLedNight

	cbi	PortB, 1
	sbi	PortB, 2

	ret

fLedOff:
	clr	i
	out	TCCR1A, i
	out	TCCR1B, i
	cbi	PortB, 1
	cbi	PortB, 2
	
	ret
	
fLedFade:
	ldi	i, 60
	mul	i, tMinute

	clr	i
	add	r0, tSecond
	adc	r1, i

	ldi	i, (0b10<<WGM10)|(0b10<<COM1A0)|(0b11<<COM1B0)
	out	TCCR1A, i
	ldi	i, (0b11<<WGM12)|(1<<CS10)
	out	TCCR1B, i
	ldi	i, 0x0E
	out	ICR1H, i
	ldi	i, 0x10
	out	ICR1L, i

	out	OCR1AH, r1
	out	OCR1AL, r0
	out	OCR1BH, r1
	out	OCR1BL, r0

	ret

      
      .org 0x0200
tableMonth:                                     ;Wrap day after last day of month
      .db   31+1, 77, 31+1, 30+1, 31+1, 30+1, 31+1, 31+1, 30+1, 31+1, 30+1 ,31+1
